import json
import logging
import uuid

import psycopg2
import urllib3

import yaml
from django.conf import settings
from django.contrib.auth.models import User
from django.utils import timezone
from kubernetes import client
from kubernetes.client import CoreV1Api, ApiException
from kubernetes.stream import stream
from kubernetes.watch import watch
from kubernetes.watch.watch import SimpleNamespace

from clusterCore.exceptions import *
from clusterCore.api.serializer import c123_HostSerializer, object_to_json, c123_inventory_serializer
from clusterCore.exceptions import (
    APIException404,
    APIException401,
    APIException500,
    APIException409,
    APIException406,
    APIException400,
    APIException410,
)
from clusterCore.models import (
    Task,
    c123_node,
    ClusterSizing,
    SystemConfig,
    c123_inventory, c123_database, c123_host, IntegratedService,
)
from django_apscheduler.util import close_old_connections
from clusterCore.enums import DBType, ClusterRoles, Status, Inventory, Schedule, TaskState, Timeout, Severity, \
    LicenseType
from clusterCore.tasks import (
    generate_license_summary,
    revoke_license,
    spread_licenses_by_type,
    cleanup_logs,
    update_host_usage,
    remote_command,
    allow_or_403, create_ssh_connection, check_host_connection, local_command, connect_to_sqlserver_instance,
)

TASK_CERTIFICATE_FILES_DIR = "/123cluster/certs"
SHELL_SAFE_TRANSFORM = str.maketrans({":": "_", ".": "_"})

unlicense_notice = """
Dear {{user}}

We would like to inform you that the cluster {{cluster}} is currently having nodes without full license. To ensure uninterrupted service and access to all functionalities, we kindly request you to acquire and apply the appropriate license {{within_time}}.
Nodes that require attention:
{{nodes}}
If you have any questions or need further assistance regarding licensing, please don't hesitate to contact our support team.

Thank you for your attention to this matter.

Best regards,
{{company}}
"""


# Disable only InsecureRequestWarning
urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)


def get_core_v1_api(ip: str, username: str, password: str, private_key: str, ssh_port: int = 22) -> CoreV1Api | None:
    """
    Establish a connection to a Kubernetes cluster, retrieve or generate a bearer token,
    and return a configured CoreV1Api instance for interacting with the cluster.

    Args:
        ip (str): The IP address of the Kubernetes master node.
        username (str): The SSH username for connecting to the Kubernetes node.
        password (str): The SSH password for authentication (if not using a private key).
        private_key (str): The private key for SSH authentication (if password is not used).
        ssh_port (int, optional): The SSH port to connect to the Kubernetes node. Defaults to 22.

    Returns:
        CoreV1Api | None: An instance of Kubernetes CoreV1Api if successful, otherwise None.

    Example Usage:
        ```python
        v1 = get_core_v1_api(
            ip="192.168.1.100",
            username="root",
            password="password",
            private_key="/path/to/key.pem"
        )
        pods = v1.list_namespaced_pod(namespace="postgres")
        print(pods)
        ```
    """
    ssh_client = create_ssh_connection(ip, username, password, private_key, ssh_port)

    # check whether the token already exists
    jsn = json.loads(
        remote_command(ssh_client, f'kubectl get secret 123cluster-token-access-secret -o jsonpath="{{.data.token}}" | base64 --decode', True),
        strict=False,
    )
    if "not found" in jsn["stderr"].lower():
        # if the token does not exist, apply the manifest to create one
        service_account = {"apiVersion": "v1", "kind": "ServiceAccount", "metadata": {"name": "123cluster-token-access"}}

        cluster_role_binding = {
            "apiVersion": "rbac.authorization.k8s.io/v1",
            "kind": "ClusterRoleBinding",
            "metadata": {"name": "123cluster-token-access-binding"},
            "subjects": [{"kind": "ServiceAccount", "name": "123cluster-token-access", "namespace": "default"}],
            "roleRef": {
                "kind": "ClusterRole",
                "name": "cluster-admin",  # grant full access
                "apiGroup": "rbac.authorization.k8s.io",
            },
        }

        secret = {
            "apiVersion": "v1",
            "kind": "Secret",
            "metadata": {
                "name": "123cluster-token-access-secret",
                "annotations": {"kubernetes.io/service-account.name": "123cluster-token-access"},
            },
            "type": "kubernetes.io/service-account-token",
        }

        service_account = yaml.dump(service_account, default_flow_style=False)
        cluster_role_binding = yaml.dump(cluster_role_binding, default_flow_style=False)
        secret = yaml.dump(secret, default_flow_style=False)

        remote_command(ssh_client, f"kubectl apply -f - <<EOF\n{service_account}\nEOF", True)
        remote_command(ssh_client, f"kubectl apply -f - <<EOF\n{cluster_role_binding}\nEOF", True)
        remote_command(ssh_client, f"kubectl apply -f - <<EOF\n{secret}\nEOF", True)

        jsn = json.loads(
            remote_command(ssh_client, f'kubectl get secret 123cluster-token-access-secret -o jsonpath="{{.data.token}}" | base64 --decode', True),
            strict=False,
        )

    # configure authorization
    aToken = jsn["stdout"]
    jsn = json.loads(
        remote_command(ssh_client, "kubectl config view --minify -o jsonpath='{.clusters[0].cluster.server}' | awk -F[:] '{print $NF}'", True),
        strict=False,
    )
    api_port = jsn["stdout"]

    aConfiguration = client.Configuration()
    aConfiguration.host = f"https://{ip}:{api_port}"
    # TODO: Configure SSL certificates
    # Note: Kubernetes certificates are self-signed.
    #       The method to obtain certificates varies between Kubernetes versions.
    aConfiguration.verify_ssl = False
    aConfiguration.api_key = {"authorization": "Bearer " + aToken}
    aApiClient = client.ApiClient(aConfiguration)

    return client.CoreV1Api(aApiClient)


def node_exec_command(v1: CoreV1Api, node_name: str, command: str) -> dict[str, str]:
    """
    Executes a shell command on a specified Kubernetes node by creating a privileged pod in default namespace.

    This function dynamically creates a temporary privileged pod on the target node,
    executes the given command inside the pod, and then ensures the pod is deleted
    after execution.

    Args:
        v1 (CoreV1Api): The Kubernetes CoreV1Api client instance used to interact with the cluster.
        node_name (str): The name of the Kubernetes node where the command should be executed.
        command (str): The shell command to run inside the privileged pod.
    Returns:
        dict[str, str]: A dictionary containing the command output, with keys such as 'stdout' and 'stderr'.

    Note:
        - **Performance Consideration**: The function takes more time to execute as it involves pod creation and deletion.
          **If possible, consider alternative methods.**
    """
    # avoid name conflicts when executing two commands simultaneously
    pod_name = f"{node_name}-ssh-{uuid.uuid4().hex[:5]}"
    namespace = "default"

    pod_manifest = {
        "apiVersion": "v1",
        "kind": "Pod",
        "metadata": {"name": pod_name, "namespace": namespace},
        "spec": {
            "hostPID": True,
            "hostIPC": True,
            "hostNetwork": True,
            # node where to schedule the pod
            "nodeName": node_name,
            "containers": [
                {
                    "name": "executor",
                    "image": "alpine:latest",
                    "command": ["sleep", "infinity"],
                    "securityContext": {"privileged": True},
                    "volumeMounts": [{"name": f"{node_name}-filesystem", "mountPath": "/"}],
                }
            ],
            "volumes": [{"name": f"{node_name}-filesystem", "hostPath": {"path": "/", "type": "Directory"}}],
        },
    }

    logging.info(f"Creating pod {pod_name} on node {node_name}...")

    try:
        v1.create_namespaced_pod(body=pod_manifest, namespace=namespace)

        count = 10
        w = watch.Watch()
        for event in w.stream(v1.list_namespaced_pod, namespace=namespace, field_selector=f"metadata.name={pod_name}", timeout_seconds=10):
            pod = event["object"]
            if pod.status.phase == "Running":
                w.stop()
            count -= 1
            if not count:
                w.stop()

        print(f"node_exec_command in node {node_name} {format(command)}")

        exec_command = ["/bin/sh", "-c", command]
        resp = stream(
            v1.connect_get_namespaced_pod_exec,
            pod_name,
            namespace,
            command=exec_command,
            stderr=True,
            stdin=False,
            stdout=True,
            tty=False,
            _preload_content=False,
        )

        stdout = ""
        stderr = ""

        while resp.is_open():
            resp.update(timeout=1)
            if resp.peek_stdout():
                stdout += resp.read_stdout()
            if resp.peek_stderr():
                stderr += resp.read_stderr()

        resp.close()

        return {"stdout": stdout, "stderr": stderr}
    finally:
        # always delete the pod, even if an exception occurs
        v1.delete_namespaced_pod(pod_name, namespace, body=client.V1DeleteOptions())


def get_unused_node_ip(v1: CoreV1Api, namespace: str, cluster_label: str) -> str | None:
    """
    Retrieve the IP address of an unused worker node in a Kubernetes cluster.

    This function scans all nodes in the cluster and determines which ones are already in use by pods or
    persistent volumes associated with the given `cluster_label`. It returns the internal IP address of
    an available node that is not designated as a master or control-plane node.

    Args:
        v1 (CoreV1Api): An instance of the Kubernetes CoreV1Api client.
        namespace (str): The namespace in which to check for running pods.
        cluster_label (str): The label used to identify resources belonging to a specific cluster.

    Returns:
        str | None: The internal IP address of an available worker node, or `None` if all nodes are in use.

    Example:
        ```python
        node_ip = get_unused_node_ip(
            v1=v1,
            namespace="postgres",
            cluster_label="123cluster-postgres"
        )

        if node_ip:
            print(f"Available node IP: {node_ip}")
        else:
            print("No available worker nodes.")
        ```
    """
    nodes = v1.list_node().items
    pods = v1.list_namespaced_pod(namespace, label_selector=f"cluster={cluster_label}").items
    pv_list = v1.list_persistent_volume(label_selector=f"cluster={cluster_label}").items

    used_nodes = {pod.spec.node_name for pod in pods if pod.spec.node_name} | {
        pv.spec.node_affinity.required.node_selector_terms[0].match_expressions[0].values[0]
        for pv in pv_list
        if pv.spec.node_affinity and pv.spec.node_affinity.required
    }

    for node in nodes:
        node_name = node.metadata.name

        # Exclude master nodes
        if "node-role.kubernetes.io/master" in node.metadata.labels or "node-role.kubernetes.io/control-plane" in node.metadata.labels:
            continue

        if node_name not in used_nodes:
            # Extract the node's IP
            for addr in node.status.addresses:
                if addr.type == "InternalIP":
                    return addr.address

    # If all nodes are used
    return None


def is_path_exists_on_node(cp: SimpleNamespace, node_name: str, path: str) -> bool:
    """
    Checks if a given path exists on a specified node.

    Args:
        cp (SimpleNamespace): A namespace object containing control plane SSH connection details.
        node_name (str): The name of the node where the command will be executed.
        path (str): The file or directory path to check.

    Returns:
        bool: True if the path exists, False if it doesn't

    Raises:
        APIException401: If there is an issue with SSH authentication.
        APIException500: If an unexpected error occurs while executing the command.
    """
    ssh_client = create_ssh_connection(cp.ip, cp.username, cp.password, cp.private_key, cp.ssh_port)
    if isinstance(ssh_client, tuple):
        if hasattr(ssh_client[1], "strerror"):
            raise APIException401(client[1].strerror)
        raise APIException401(client[1])

    v1 = get_core_v1_api(cp.ip, cp.username, cp.password, cp.private_key, cp.ssh_port)

    resp = node_exec_command(v1, node_name, f"sudo stat '{path}'")
    if resp["stderr"]:
        if "no such file or directory" in resp["stderr"].lower():
            return False
        else:
            raise APIException500(resp["stderr"])

    return True


def is_port_in_use_on_node(cp: SimpleNamespace, node_name: str, port: int) -> bool:
    """
    Checks if a given port in use on a specified node.

    Args:
        cp (SimpleNamespace): A namespace object containing control plane SSH connection details.
        node_name (str): The name of the node where the command will be executed.
        port (int): The port to check.

    Returns:
        bool: True if the port in use, False if it doesn't

    Raises:
        APIException401: If there is an issue with SSH authentication.
        APIException500: If an unexpected error occurs while executing the command.
    """
    ssh_client = create_ssh_connection(cp.ip, cp.username, cp.password, cp.private_key, cp.ssh_port)
    if isinstance(ssh_client, tuple):
        if hasattr(ssh_client[1], "strerror"):
            raise APIException401(client[1].strerror)
        raise APIException401(client[1])

    v1 = get_core_v1_api(cp.ip, cp.username, cp.password, cp.private_key, cp.ssh_port)

    resp = node_exec_command(v1, node_name, f"sudo netstat -tunlp|grep -w '{port}'")
    if resp["stdout"]:
        return True

    if resp["stderr"] and "warning" not in resp["stderr"].lower():
        raise APIException500(resp["stderr"])

    return False


def check_space(node, path, source_path = None, required_space = None):
    app_host = c123_host.objects.filter(ex_params__myself = True).first()
    if not app_host:
        raise APIException404("could not determine the app host")

    jsn = json.loads(local_command(f"sudo stat '{path}'", True), strict = False)
    if jsn['stderr'] != '':
        if "cannot stat" in jsn["stderr"].lower():
            jsn = json.loads(local_command(f"sudo mkdir -p '{path}'", True), strict=False)
            error = str(jsn['stderr'])
            if error != '':
                raise APIException400(error)

    cmd = f'''df -k "{path}" | awk 'NR>1 {{print $4}}' '''
    jsn = json.loads(local_command(cmd, True), strict=False)

    if jsn['stderr'] != '':
            raise APIException400(jsn['stderr'])

    available_space = jsn['stdout']


    # case when backup storage have different os family than node machine
    if not required_space:
        match node.type:
            case DBType.MSSQL.value:
                conn = connect_to_sqlserver_instance(node)
                if isinstance(conn, tuple):
                    if hasattr(conn[1], 'strerror'):
                        raise APIException401(conn[1].strerror)
                    raise APIException401(str(conn[1]))
                else:
                    try:
                        with conn:
                            with conn.cursor() as cur:
                                cur.execute("SELECT SUM(mf.size * 8) AS SizeKB FROM sys.master_files mf JOIN sys.databases d ON mf.database_id = d.database_id WHERE mf.database_id > 4 GROUP BY mf.database_id")
                                res = cur.fetchone()
                                if not res or not res[0]:
                                    raise APIException500("Could not obtain required space for backup.")
                                required_space = float(res[0])
                    except Exception as error:
                        raise APIException500(str(error))
            case _ :
                if source_path:
                    client = create_ssh_connection(node.host.name, node.host.username, node.host.password, node.host.private_key, node.host.ssh_port)
                    if isinstance(client, tuple):
                        if hasattr(client[1], 'strerror'):
                            raise APIException401(client[1].strerror)
                        raise APIException401(client[1])
                    jsn = json.loads(remote_command(client, f"sudo du -s -BK '{source_path}' | awk '{{print $1}}' | tr -d 'K'", True), strict = False)
                    if jsn['stderr'] != '':
                        raise APIException400(jsn['stderr'])
                    required_space = float(jsn['stdout'])

    if not source_path:
        if float(available_space) > float(required_space):
            return {'available_space': available_space, 'status': 200}
        else:
            return {'available_space': available_space, 'status': 406}
    else:
        if float(available_space) > float(required_space):
            return {'available_space': available_space, 'required_space': required_space, 'status': 200}
        else:
            raise APIException406(f'{path} has {available_space} KB free space. {required_space} KB required')


@close_old_connections
def drop_k8s_pg_cluster(cluster, final_backup = False):
    cluster_pods = set(c123_node.objects.filter(cluster=cluster.pk, ind_deleted=False))
    cluster_hosts = set(pod.host for pod in cluster_pods)
    primary_pod = next((p for p in cluster_pods if p.cluster_role == ClusterRoles.PRIMARY.value), None)
    kubernetes_api_data = IntegratedService.objects.filter(pk=primary_pod.ex_params["service_id"]).first()
    node = c123_node.objects.filter(cluster=cluster.pk, cluster_role=ClusterRoles.PRIMARY.value, type = DBType.POSTGRES_K8S.value).first()
    node_data = ''
    host_data = ''

    if final_backup:
        node_data = object_to_json(node)
        host_data = c123_HostSerializer(node.host).data
        app_host = c123_host.objects.filter(ex_params__myself=True).first()
        if not app_host:
            raise APIException404("could not determine the app host")
        backup_host = c123_HostSerializer(app_host).data
        size_scheme, created = ClusterSizing.objects.get_or_create(db_type = DBType.POSTGRES.value)
        keep_days = getattr(size_scheme, f'{node.sizing.lower()}_backup_keep_days', None)
        # create a unique string with the date
        now = timezone.now().strftime('%d%m%Y%H%M%S')
        config = SystemConfig.objects.filter(db_type = DBType.POSTGRES.value).first()
        export_path = config.exports_path
        ex_params = {
            'size': 0,
            'duration': 0,
            'plain': False,
            'belong_to': DBType.POSTGRES_K8S.value,
            'node_version': node.version,
            'compress': True,
            'backup_type': 'dumpall',
            'modified_date': '',
            'backup_path': f'{export_path}/dumps/{node.host.name}_{node.port}/{now}',
        }

        final_backup = c123_inventory.objects.create(name = "{}_{}".format(node.name, now),
        node = node,
        type = Inventory.EXPORT.value,
        created_by = node.created_by,
        is_sched = False,
        keep_days = keep_days,
        ex_params = ex_params
        )

    for p in cluster_pods:
        cleanup_logs(p.pk, p.type)
        update_host_usage(p.host)
        p.status = Status.DROPPING.value
        p.save()

        pod_jobs = [job for job in settings.scheduler.get_jobs() if p.name in job.id]
        for job in pod_jobs:
            job.remove()



    tmp_data = {
        "created_by": cluster.created_by.id,
        "kubernetes_api_data": object_to_json(kubernetes_api_data),
        "cluster_data": object_to_json(cluster),
        'node_data': node_data,
        'host_data': host_data,
        'final_backup': c123_inventory_serializer(final_backup).data if final_backup else final_backup,
        "cluster_hosts_data": [object_to_json(h) for h in cluster_hosts],
        "cluster_pods_data": [object_to_json(p) for p in cluster_pods]
    }

    task = Task.objects.create(
        task="drop_k8s_pg_cluster_task",
        displayment=f"drop cluster {cluster.name}",
        target_type=DBType.POSTGRES_K8S.value,
        target_id=cluster.id,
        # cluster has no attribute 'org_id'
        org_id=primary_pod.org_id,
        schedule=Schedule.NOW.value,
        status=TaskState.PENDING.value,
        parameters=json.dumps(tmp_data),
        created_by=cluster.created_by,
    )

    licence_summary = generate_license_summary()
    for p in cluster_pods: revoke_license(p)
    spread_licenses_by_type(LicenseType.POSTGRES.value, licence_summary.postgresql)

    return task


@close_old_connections
def drop_k8s_pg_pod(pod, is_make_final_backup=False):
    kubernetes_api_data_id = pod.ex_params["service_id"]
    kubernetes_api_data = IntegratedService.objects.filter(pk=kubernetes_api_data_id).first()

    cleanup_logs(pod.pk, pod.type)
    update_host_usage(pod.host)

    pod.status = Status.DROPPING.value
    pod.save()

    pod_jobs = [job for job in settings.scheduler.get_jobs() if pod.name in job.id]
    for job in pod_jobs:
        job.remove()

    pod.cluster.ex_params["instances"] -= 1
    pod.cluster.save()

    tmp_data = {
        "created_by": pod.created_by.id,
        "kubernetes_api_data": object_to_json(kubernetes_api_data),
        "cluster_data": object_to_json(pod.cluster),
        "pod_data": object_to_json(pod)
    }

    task = Task.objects.create(
        task="drop_k8s_pg_pod_task",
        displayment=f"drop {pod.cluster_role.lower()} {pod.name}",
        target_type=DBType.POSTGRES_K8S.value,
        target_id=pod.id,
        org_id=pod.org_id,
        schedule=Schedule.NOW.value,
        status=TaskState.PENDING.value,
        parameters=json.dumps(tmp_data),
        created_by=pod.created_by,
    )

    licence_summary = generate_license_summary()
    revoke_license(pod)
    spread_licenses_by_type(LicenseType.POSTGRES.value, licence_summary.postgresql)

    return task


def drop_mongo_k8s_pod(node, final_backup=False):
    existing_drop = Task.objects.filter(
        target_id=node.id, task__in=["uninstall_mongodb_k8s_standalone_task"], status__in=[TaskState.PENDING.value, TaskState.RUNNING.value]
    )
    if existing_drop:
        raise APIException409(f"Already runs drop node for cluster {node.cluster.name}")

    node.status = Status.DROPPING.value
    node.save()

    cleanup_logs(node.pk, node.type)
    update_host_usage(node.host)

    node_data = object_to_json(node)
    host_data = c123_HostSerializer(node.host).data
    created_by = node.created_by
    tmp_data = {"node_data": node_data, "host_data": host_data, "created_by": created_by.id}
    tmp_data = json.dumps(tmp_data)
    summary = generate_license_summary()
    node.ind_license = False
    node.save()

    from clusterCore.api.licensing.views import LicenseType

    revoke_license(node)
    spread_licenses_by_type(LicenseType.MONGO.value, summary.mongo)

    task = Task.objects.create(
        task="uninstall_mongodb_k8s_standalone_task",
        displayment=f"drop mongo {node.cluster_role.lower()} {node.name}",
        target_type=DBType.MONGO_K8S.value,
        target_id=node.id,
        org_id=node.org_id,
        schedule=Schedule.NOW.value,
        status=TaskState.PENDING.value,
        parameters=tmp_data,
        created_by=created_by,
    )
    return task


def export_mongo_k8s_node(data):
    try:

        node_id = data.get("node_id")
        compress = data.get("compress")
        include_data = data.get("include_data")
        path = data.get("export_path")
        dump_name = data.get("dump_name")
        api_call = data.get("rest_api")
        created_by = User.objects.filter(id=data.get("created_by")).first()

        serializer = c123_inventory_serializer(data={"name": ""}, partial=True)
        if serializer.is_valid():
            if api_call:
                mongo_node = c123_node.objects.filter(name=node_id, type=DBType.MONGO_K8S.value, ind_deleted=False).first()
            else:
                mongo_node = c123_node.objects.filter(pk=node_id, type=DBType.MONGO_K8S.value, ind_deleted=False).first()
            if not mongo_node:
                raise APIException406(f"node id {node_id} not available")
            allow_or_403(mongo_node)
            size_scheme, created = ClusterSizing.objects.get_or_create(db_type=DBType.MONGO_K8S.value)
            keep_days = getattr(size_scheme, f"{mongo_node.sizing.lower()}_export_keep_days", None)
            dump = serializer.create(serializer.validated_data)
            dump.name = dump_name
            dump.node = mongo_node
            if keep_days:
                dump.keep_days = keep_days
            dump.type = Inventory.EXPORT.value
            dump.created_by = created_by
            dump.ex_params = {"size": 0}
            dump.ex_params["belong_to"] = DBType.MONGO_K8S.value
            dump.ex_params["node_version"] = mongo_node.version
            dump.ex_params["duration"] = 0
            dump.ex_params["compress"] = compress
            dump.ex_params["modified_date"] = ""
            # create a unique string with the date
            now = timezone.now().strftime("%d%m%Y%H%M%S")
            dump.name = "{}_{}".format(dump.name, now)
            dump.save()
            mongo_node.status = Status.BACKUP.value
            mongo_node.save()
            # serialize objects for the tasks
            node_data = object_to_json(mongo_node)
            host_data = c123_HostSerializer(mongo_node.host).data
            # backup_host = c123_HostSerializer(app_host).data
            dump.ex_params["backup_path"] = "/123cluster-test/mongodb/168.119.125.84:27017/data/dump/"
            dump.save()
            tmp_data = json.dumps(
                {
                    "db_name": None,
                    "node_data": node_data,
                    "host_data": host_data,
                    "dump_data": object_to_json(dump),
                    "compress": compress,
                    "include_data": include_data,
                    "created_by": created_by.id,
                }
            )
            task = Task.objects.create(
                task="mongodump_task",
                displayment=f"dump mongo node {mongo_node.name}",
                target_type=DBType.MONGO_K8S.value,
                target_id=mongo_node.id,
                org_id=mongo_node.org_id,
                schedule=Schedule.NOW.value,
                status=TaskState.PENDING.value,
                parameters=tmp_data,
                created_by=created_by,
            )
            return task

        else:
            raise APIException400(f"{serializer.errors}")

    except Exception as e:
        raise APIException500(str(e))


def export_k8s_pg_node(data):
    try:
        node_id = data.get("node_id")
        compress = data.get("compress")
        include_data = data.get("include_data")
        include_publications = data.get("include_publications")
        include_subscriptions = data.get("include_subscriptions")
        use_inserts = data.get("use_inserts")
        path = data.get("export_path")
        export_name = data.get("export_name")
        api_call = data.get("rest_api")
        created_by = User.objects.filter(id=data.get("created_by")).first()
        app_host = c123_host.objects.filter(ex_params__myself=True).first()
        if not app_host:
            raise APIException404("could not determine the app host")
        serializer = c123_inventory_serializer(data={"name": ""}, partial=True)
        if serializer.is_valid():
            if api_call:
                pg_node = c123_node.objects.filter(type=DBType.POSTGRES_K8S.value, name=node_id, ind_deleted=False).first()
            else:
                pg_node = c123_node.objects.filter(type=DBType.POSTGRES_K8S.value, pk=node_id, ind_deleted=False).first()
            if not pg_node:
                raise APIException410("node id {} not available".format(node_id))
            allow_or_403(pg_node)
            size_scheme, created = ClusterSizing.objects.get_or_create(db_type=DBType.POSTGRES.value)
            keep_days = getattr(size_scheme, f"{pg_node.sizing.lower()}_export_keep_days", None)
            export = serializer.create(serializer.validated_data)
            export.name = export_name
            export.node = pg_node
            if keep_days:
                export.keep_days = keep_days
            export.type = Inventory.EXPORT.value
            export.created_by = created_by
            export.ex_params = {"size": 0}
            export.ex_params["belong_to"] = DBType.POSTGRES_K8S.value
            export.ex_params["node_version"] = pg_node.version
            export.ex_params["backup_type"] = "dumpall"
            export.ex_params["duration"] = 0
            export.ex_params["modified_date"] = ""
            # create a unique string with the date
            now = timezone.now().strftime("%d%m%Y%H%M%S")
            export.name = "{}_{}".format(export.name, now)
            export.save()
            pg_node.status = Status.BACKUP.value
            pg_node.save()
            # serialize objects for the tasks
            node_data = object_to_json(pg_node)
            host_data = c123_HostSerializer(pg_node.host).data
            export.ex_params["backup_path"] = f"{path}/dumps/{pg_node.host.name}_{pg_node.port}/{now}"
            export.ex_params["compress"] = compress
            export.save()
            backup_host = c123_HostSerializer(app_host).data
            tmp_data = json.dumps(
                {
                   'node_data': node_data,
                   'host_data': host_data,
                   'backup_host': backup_host,
                   'backup_data': object_to_json(export),
                   'backup_type':'dumpall',
                   'include_data': include_data,
                   'include_publications': include_publications,
                   'include_subscriptions': include_subscriptions,
                   'use_inserts': use_inserts,
                   'created_by': created_by.id
                }
            )
            task = Task.objects.create(
                task="pg_dump_backup_task",
                displayment="backup pg {}".format(pg_node.name),
                target_type=DBType.POSTGRES_K8S.value,
                target_id=pg_node.id,
                org_id=pg_node.org_id,
                schedule=Schedule.NOW.value,
                status=TaskState.PENDING.value,
                parameters=tmp_data,
                created_by=created_by,
            )

            export.ex_params["task_id"] = task.pk
            export.save()
            return task

        else:
            raise APIException400(f"{serializer.errors}")
    except Exception as e:
        raise APIException500(str(e))

def export_k8s_pg_database(data):

    try:

        path = data.get("export_path")
        db_id = data.get('db_id')
        node_id = data.get('node_id')
        compress = data.get('compress')
        include_data = data.get('include_data')
        include_publications = data.get('include_publications')
        include_subscriptions = data.get('include_subscriptions')
        use_inserts = data.get('use_inserts')
        plain = data.get('plain')
        export_name = data.get("export_name")
        api_call = data.get('rest_api')
        created_by = User.objects.filter(id = data.get('created_by')).first()
        app_host = c123_host.objects.filter(ex_params__myself = True).first()
        if not app_host:
            raise APIException404("could not determine the app host")
        serializer = c123_inventory_serializer(data = {'name': ''}, partial = True)
        if serializer.is_valid():
            if api_call:
                pg_db = c123_database.objects.filter(type = DBType.POSTGRES_K8S.value, name=db_id, node__name = node_id, ind_deleted = False).first()
            else:
                pg_db = c123_database.objects.filter(type = DBType.POSTGRES_K8S.value, pk=db_id, ind_deleted = False).first()
            if not pg_db:
                raise APIException410("database id {} not available".format(db_id))
            pg_node = pg_db.node
            allow_or_403(pg_node)
            size_scheme, created = ClusterSizing.objects.get_or_create(db_type = DBType.POSTGRES.value)
            keep_days = getattr(size_scheme, f'{pg_node.sizing.lower()}_export_keep_days', None)
            export = serializer.create(serializer.validated_data)
            export.name = export_name
            export.database = pg_db
            export.node = pg_node
            if keep_days:
                export.keep_days = keep_days
            export.type = Inventory.EXPORT.value
            export.created_by = created_by
            export.ex_params = {"size": 0}
            export.ex_params["belong_to"] = DBType.POSTGRES_K8S.value
            export.ex_params["node_version"] = pg_node.version
            export.ex_params["compress"] = compress
            export.ex_params["plain"] = plain
            export.ex_params["duration"] = 0
            export.ex_params["modified_date"] = ''
            export.ex_params["host_id"] = pg_node.host.id
            # # create a unique string with the date
            now = timezone.now().strftime('%d%m%Y%H%M%S')
            export.name = "{}_{}".format(export_name, now)
            # export.name = export_name
            export.save()
            pg_db.status = Status.BACKUP.value
            pg_db.save()
            # serialize objects for the tasks
            db_data = object_to_json(pg_db)
            host_data = c123_HostSerializer(pg_node.host).data
            backup_host = c123_HostSerializer(app_host).data
            export.ex_params["compress"] = compress
            export.ex_params["backup_path"] = f'{path}/dumps/{pg_node.host.name}_{pg_node.port}/{now}'
            export.save()
            tmp_data = json.dumps({'db_data': db_data,
                                   'node_data': object_to_json(pg_node),
                                   'backup_data': object_to_json(export),
                                   'host_data': host_data,
                                   'backup_host':backup_host,
                                   'backup_type':'dump',
                                   'include_data': include_data,
                                   'include_publications': include_publications,
                                   'include_subscriptions': include_subscriptions,
                                   'use_inserts': use_inserts,
                                   'created_by': created_by.id})
            task = Task.objects.create(task='pg_dump_backup_task',
                                                displayment='backup {}'.format(pg_db.name),
                                                target_type=DBType.POSTGRES_K8S.value,
                                                target_id=pg_db.id,
                                                org_id=pg_node.org_id,
                                                schedule=Schedule.NOW.value,
                                                status=TaskState.PENDING.value,
                                                parameters=tmp_data,
                                                created_by=created_by)
            export.ex_params['task_id'] = task.pk
            export.save()
            return task
        else:
            raise APIException400(f"{serializer.errors}")
    except Exception as e:
            raise APIException500(str(e))

def checkK8SPathExists(path, host_type = None) -> bool:
    """check if a directory exists on remote host"""
    if host_type == 'windows':

        jsn = json.loads(local_command(f'dir "{path}"', True), strict=False)
        if jsn['stderr'] :
            if any(err in jsn['stderr'].lower() for err in ["cannot find", "not found"]):
                return False

            raise APIException500(jsn['stderr'])

        return True
    else:

        jsn = json.loads(local_command(f"sudo stat '{path}'", True), strict = False )
        if jsn['stderr'] != '':
            if "cannot stat" in jsn["stderr"].lower():
                return False
            else:
                raise APIException500(jsn['stderr'])

        return True


def get_k8s_postgres_connection(node, db_name="postgres"):
    try:

        result = check_host_connection(node)

        global TASK_CERTIFICATE_FILES_DIR
        conn_params = {
                'dbname': db_name,
                'user':  node.database_user,
                'password':  node.database_pwd,
                'host': node.name.split(':')[0],
                'port': node.port,
                'connect_timeout': Timeout.BACKEND_POSTG.value
        }
        certificate_authentication = node.ex_params.get('certificate_authentication')
        if certificate_authentication:
            certificate_root = f"{TASK_CERTIFICATE_FILES_DIR}/{conn_params['host']}_{conn_params['port']}"
            conn_params['sslmode'] = 'verify-full'
            conn_params['sslrootcert'] = f"{certificate_root}/root.crt"
            conn_params['sslcert'] = f"{certificate_root}/client.crt"
            conn_params['sslkey'] = f"{certificate_root}/client.key"
        conn = psycopg2.connect(**conn_params)
        node.severity = Severity.OK.value
        node.status = Status.OK.value
        node.ind_started = True
        node.save()
        c123_database.objects.filter(node = node, ind_deleted = False).update(severity = Severity.OK.value)
        return conn
    except Exception as e:
        if not isinstance(result, tuple):
            if node.status in [Status.STOPPING.value, Status.STOPPED.value]:
                node.status = Status.STOPPED.value
            else:
                node.status = Status.DOWN.value

            node.severity = Severity.NOT_OK.value
            node.ind_started = False
            node.save()

            c123_database.objects.filter(node = node, ind_deleted = False).update(severity = Severity.NOT_OK.value)
        logging.error(f"get_postgres_connection: {e}")
        return (False, e)
